#!/usr/bin/python3
import math
import sys


def dbg(*args):
    print(*args, file=sys.stderr)


primes = (
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
    101,
    103,
    107,
    109,
    113,
    127,
    131,
    137,
    139,
    149,
)


def is_prime(n):
    if n in primes:
        return True
    if n % 2 == 0:
        return False
    s = int(math.sqrt(n))
    for p in primes:
        if p > s:
            return True
        if n % p == 0:
            return False
    raise ValueError("Too big n=%d" % (n,))


def print_pprods_h(prods, factors):
    tab = "    "
    tmpl = """
#ifndef PPRODS_H
#define PPRODS_H

// This file is auto-generated by gen_pprods.py

// The array contains products of consecutive odd prime numbers grouped so that
// each product fits into 64 bits.
static const uint64_t pprods[] = {
%s
};
#endif // PPRODS_H
"""
    tmpl = tmpl.strip()
    prods_str = ""
    for i, pr in enumerate(prods):
        factors_str = "*".join(map(str, factors[i]))
        extra_spaces = " " * (16 - (pr.bit_length() + 3) // 4)
        prods_str += tab + "0x%x,  %s// %s\n" % (pr, extra_spaces, factors_str)
    print(tmpl % (prods_str.rstrip(),))


n = int(sys.argv[1])
cnt = 0
pr = 1
arr = []
prods = []
factors = []
for i in range(3, n):
    if is_prime(i):
        # print(i, '', end='')
        cnt += 1
        pr *= i
        if pr.bit_length() > 64:
            pr //= i
            # print('%s: %s' % (hex(pr), arr))
            prods.append(pr)
            factors.append(arr)
            pr = i
            arr = []
        arr.append(i)

print_pprods_h(prods, factors)

dbg()
dbg("n_primes=%d n_prods=%d last_prime=%d" % (cnt, len(prods), factors[-1][-1]))
# dbg(*list('%s,' % (hex(i),) for i in prods))



import os
import platform
import shutil
import subprocess
import sys

from setuptools import Command, Extension, setup
from setuptools.command.build import build
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install

BUILD_HOOKS = []
INSTALL_HOOKS = []


def add_install_hook(hook):
    INSTALL_HOOKS.append(hook)


def add_build_hook(hook):
    BUILD_HOOKS.append(hook)


class HookCommand(Command):
    def __init__(self, dist):
        self.dist = dist
        Command.__init__(self, dist)

    def initialize_options(self, *args):
        self.install_dir = None
        self.build_dir = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_scripts", "build_dir"))
        self.set_undefined_options(
            "install",
            ("install_platlib", "install_dir"),
        )

    def run(self):
        for _ in self.hooks:
            _(install_dir=self.install_dir, build_dir=self.build_dir)


class build_hook(HookCommand):
    hooks = BUILD_HOOKS


class install_hook(HookCommand):
    hooks = INSTALL_HOOKS


############################################


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=["./"])
        self.sourcedir = os.path.abspath(sourcedir)


def copy_vdf_client(build_dir, install_dir):
    shutil.copy("src/vdf_client", install_dir)
    shutil.copy("src/prover_test", install_dir)
    shutil.copy("src/1weso_test", install_dir)
    shutil.copy("src/2weso_test", install_dir)


def copy_vdf_bench(build_dir, install_dir):
    shutil.copy("src/vdf_bench", install_dir)


def invoke_make(**kwargs):
    subprocess.check_output("make -C src -f Makefile.vdf-client", shell=True)


BUILD_VDF_CLIENT = os.getenv("BUILD_VDF_CLIENT", "Y") == "Y"
BUILD_VDF_BENCH = os.getenv("BUILD_VDF_BENCH", "N") == "Y"


if BUILD_VDF_CLIENT or BUILD_VDF_BENCH:
    add_build_hook(invoke_make)

if BUILD_VDF_CLIENT:
    add_install_hook(copy_vdf_client)

if BUILD_VDF_BENCH:
    add_install_hook(copy_vdf_bench)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build"
                + " the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + str(extdir),
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j", "6"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, env=env)
        subprocess.check_call(["cmake", "--build", "."] + build_args)


build.sub_commands.append(("build_hook", lambda x: True))  # type: ignore
install.sub_commands.append(("install_hook", lambda x: True))

setup(
    name="chikvdf",
    author="Florin Chirica",
    author_email="florin@chiknetwork.com",
    description="Chik vdf verification (wraps C++)",
    license="Apache-2.0",
    python_requires=">=3.9",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Chik-Network/chikvdf",
    ext_modules=[CMakeExtension("chikvdf", "src")],
    cmdclass=dict(
        build_ext=CMakeBuild, install_hook=install_hook, build_hook=build_hook
    ),
    zip_safe=False,
)



import secrets

from chikvdf import (
    create_discriminant,
    prove,
    verify_wesolowski,
    verify_n_wesolowski,
    verify_n_wesolowski_with_b,
    get_b_from_n_wesolowski,
)


def prove_n_weso(discriminant_challenge, x, discriminant_size, form_size, iters, witness, wrong_segm):
    iters_chunk = iters // (witness + 1)
    partials = []
    discriminant = create_discriminant(discriminant_challenge, discriminant_size)
    for _ in range(witness):
        result = prove(discriminant_challenge, x, discriminant_size, iters_chunk, "")
        y = result[:form_size]
        proof = result[form_size : 2 * form_size]
        partials.append((x, y, proof))
        x = y
    iters -= iters_chunk * witness
    result = prove(discriminant_challenge, x, discriminant_size, iters, "")
    y_result = result[:form_size]
    y_proof = result[form_size : 2 * form_size]
    assert verify_wesolowski(discriminant, x, y_result, y_proof, iters)
    b_hex = get_b_from_n_wesolowski(discriminant, x, y_result + y_proof, iters, 0)
    is_valid, y_from_compression = verify_n_wesolowski_with_b(
        discriminant,
        b_hex,
        x,
        y_proof,
        iters,
        0,
    )
    assert is_valid
    assert y_from_compression == y_result
    inner_proof = b""
    for x, y, proof in reversed(partials):
        b_hex = get_b_from_n_wesolowski(discriminant, x, y + proof, iters_chunk, 0)
        b = int(b_hex, 16)
        assert verify_wesolowski(discriminant, x, y, proof, iters_chunk)
        is_valid, y_from_compression = verify_n_wesolowski_with_b(
            discriminant,
            b_hex,
            x,
            proof,
            iters_chunk,
            0,
        )
        assert is_valid
        assert y == y_from_compression
        if not wrong_segm:
            inner_proof += iters_chunk.to_bytes(8, byteorder='big')
        else:
            iters_wrong = iters_chunk + 1
            inner_proof += iters_wrong.to_bytes(8, byteorder='big')
            wrong_segm = False
        inner_proof += b.to_bytes(33, byteorder='big')
        inner_proof += proof
    return y_result, y_proof + inner_proof


def test_prove_n_weso_and_verify():
    discriminant_challenge = secrets.token_bytes(10)
    discriminant_size = 512
    discriminant = create_discriminant(discriminant_challenge, discriminant_size)
    print(f"discriminant_challenge {discriminant_challenge.hex()} discriminant {discriminant}")
    form_size = 100
    initial_el = b"\x08" + (b"\x00" * 99)

    for iters in [1000000, 5000000, 10000000]:
        y, proof = prove_n_weso(discriminant_challenge, initial_el, discriminant_size, form_size, iters, 5, False)
        is_valid = verify_n_wesolowski(
            str(discriminant),
            initial_el,
            y + proof,
            iters,
            discriminant_size,
            5,
        )
        assert is_valid
        is_valid = verify_n_wesolowski(
            str(discriminant),
            initial_el,
            y + proof,
            iters + 1,
            discriminant_size,
            5,
        )
        assert not is_valid
        y, proof_wrong = prove_n_weso(discriminant_challenge, initial_el, discriminant_size, form_size, iters, 10, True)
        is_valid = verify_n_wesolowski(
            str(discriminant),
            initial_el,
            y + proof_wrong,
            iters,
            discriminant_size,
            10,
        )
        assert not is_valid
        b_hex = get_b_from_n_wesolowski(discriminant, initial_el, y + proof, iters, 5)
        is_valid, y_from_compression = verify_n_wesolowski_with_b(
            discriminant,
            b_hex,
            initial_el,
            proof,
            iters,
            5,
        )
        assert is_valid
        assert y_from_compression == y
        B = str(int(b_hex, 16))
        is_valid, y_from_compression = verify_n_wesolowski_with_b(
            discriminant,
            B,
            initial_el,
            proof,
            iters,
            5,
        )
        assert is_valid
        assert y_from_compression == y
        B_wrong = str(int(b_hex, 16) + 1)
        is_valid, y_from_compression = verify_n_wesolowski_with_b(
            discriminant,
            B_wrong,
            initial_el,
            proof,
            iters,
            5,
        )
        assert not is_valid
        assert y_from_compression == b""
        is_valid, y_from_compression = verify_n_wesolowski_with_b(
            discriminant,
            B,
            initial_el,
            proof_wrong,
            iters,
            10,
        )
        assert not is_valid
        assert y_from_compression == b""
        initial_el = y


test_prove_n_weso_and_verify()



import time

from chikvdf import verify_n_wesolowski, verify_wesolowski


class ClassGroup(tuple):
    @classmethod
    def identity_for_discriminant(class_, d):
        return class_.from_ab_discriminant(1, 1, d)

    @classmethod
    def from_ab_discriminant(class_, a, b, discriminant):
        if discriminant >= 0:
            raise ValueError("Positive discriminant.")
        if discriminant % 4 != 1:
            raise ValueError("Invalid discriminant mod 4.")
        if a == 0:
            raise ValueError("a can't be 0.")
        c = (b * b - discriminant) // (4 * a)
        p = class_((a, b, c)).reduced()
        if p.discriminant() != discriminant:
            raise ValueError("No classgroup element given the discriminant.")
        return p

    @classmethod
    def from_bytes(class_, bytearray, discriminant):
        int_size = (discriminant.bit_length() + 16) >> 4
        a = int.from_bytes(bytearray[0:int_size], "big", signed=True)
        b = int.from_bytes(bytearray[int_size:], "big", signed=True)
        return class_.from_ab_discriminant(a, b, discriminant)

    def __new__(cls, t):
        a, b, c = t
        return tuple.__new__(cls, (a, b, c))

    def __init__(self, t):
        super(ClassGroup, self).__init__()
        self._discriminant = None

    def __eq__(self, obj):
        return (
            isinstance(obj, ClassGroup)
            and obj[0] == self[0]
            and obj[1] == self[1]
            and obj[2] == self[2]
        )

    def identity(self):
        return self.identity_for_discriminant(self.discriminant())

    def discriminant(self):
        if self._discriminant is None:
            a, b, c = self
            self._discriminant = b * b - 4 * a * c
        return self._discriminant

    def reduced(self):
        a, b, c = self.normalized()
        while a > c or (a == c and b < 0):
            if c == 0:
                raise ValueError("Can't reduce the form.")
            s = (c + b) // (c + c)
            a, b, c = c, -b + 2 * s * c, c * s * s - b * s + a
        return self.__class__((a, b, c)).normalized()

    def normalized(self):
        a, b, c = self
        if -a < b <= a:
            return self
        r = (a - b) // (2 * a)
        b, c = b + 2 * r * a, a * r * r + b * r + c
        return self.__class__((a, b, c))

    def serialize(self):
        r = self.reduced()
        int_size_bits = int(self.discriminant().bit_length())
        int_size = (int_size_bits + 16) >> 4
        return b"".join(
            [x.to_bytes(int_size, "big", signed=True) for x in [r[0], r[1]]]
        )


def deserialize_proof(proof_blob, discriminant):
    int_size = (discriminant.bit_length() + 16) >> 4
    proof_arr = [
        proof_blob[_ : _ + 2 * int_size]
        for _ in range(0, len(proof_blob), 2 * int_size)
    ]
    return [ClassGroup.from_bytes(blob, discriminant) for blob in proof_arr]


def check_proof_of_time_nwesolowski(
    discriminant, x, proof_blob, iterations, int_size_bits, depth
):
    """
    Check the nested wesolowski proof. The proof blob
    includes the output of the VDF, along with the proof. The following
    table gives an example of the checks for a depth of 2.
    x   |  proof_blob
    ---------------------------------------------
    x   |  y3, proof3, y2, proof2, y1, proof1
    y1  |  y3, proof3, y2, proof2
    y2  |  y3, proof3
    """

    try:
        int_size = (int_size_bits + 16) >> 4
        if len(proof_blob) != 4 * int_size + depth * (8 + 4 * int_size):
            return False
        new_proof_blob = proof_blob[: 4 * int_size]
        iter_list = []
        for i in range(4 * int_size, len(proof_blob), 4 * int_size + 8):
            iter_list.append(int.from_bytes(proof_blob[i : (i + 8)], byteorder="big"))
            new_proof_blob = (
                new_proof_blob + proof_blob[(i + 8) : (i + 8 + 4 * int_size)]
            )
        proof_blob = new_proof_blob

        result_bytes = proof_blob[: (2 * int_size)]
        proof_bytes = proof_blob[(2 * int_size) :]
        y = ClassGroup.from_bytes(result_bytes, discriminant)

        proof = deserialize_proof(proof_bytes, discriminant)
        if depth * 2 + 1 != len(proof):
            return False

        for _ in range(depth):
            iterations_1 = iter_list[-1]
            if not verify_wesolowski(
                str(discriminant),
                str(x[0]),
                str(x[1]),
                str(proof[-2][0]),
                str(proof[-2][1]),
                str(proof[-1][0]),
                str(proof[-1][1]),
                iterations_1,
            ):
                return False
            x = proof[-2]
            iterations = iterations - iterations_1
            proof = proof[:-2]
            iter_list = iter_list[:-1]

        return verify_wesolowski(
            str(discriminant),
            str(x[0]),
            str(x[1]),
            str(y[0]),
            str(y[1]),
            str(proof[-1][0]),
            str(proof[-1][1]),
            iterations,
        )
    except Exception:
        return False


def oldie(disc, a, b, proof, iter, sb, witness, doold):
    if z == 0:
        x = ClassGroup.from_ab_discriminant(int(a), int(b), int(disc))

        return check_proof_of_time_nwesolowski(
            int(disc),
            x,
            proof,
            iter,
            sb,
            witness,
        )

    return verify_n_wesolowski(
        disc,
        a,
        b,
        proof,
        iter,
        sb,
        witness,
    )


for z in range(2):
    iters = 1000
    t1 = time.time()
    for i in range(iters):
        is_valid = oldie(
            str(
                -131653324254138636653163861414331698305531090221496467927360326686715180966094250598321899621249972220387687148397451395672779897144571112116763666653213748473909547482437246405018707472153290116227072825447643324530509016778432769802300913461285128339119844239772697652504835780459732685000796733645621728639
            ),
            str(2),
            str(1),
            bytes.fromhex(
                "003f360be667de706fe886f766fe20240de04fe2c2f91207f1bbdddf20c554ab8d168b2ce9664d75f4613375a0ab12bf8158983574c9f5cd61c6b8a905fd3fa6bbffc5401b4ccedbe093b560293263a226e46302e720726586251116bc689ef09dc70d99e0a090c4409f928e218e85032fdbee02fedd563073be555b75a70a2d6a430033bc7a4926e3504e87698a0ace0dee6364cced2e9142b4e4cbe55a6371aab41e501ceed21d79d3a0dbbd82ce913c5de40b13eb7c59b1b52b6ef270ee603bd5e7fffcc9f5fae6dbd5aeec394181af130c0fdd195b22be745449b7a584ac80fc75ed49acfdb4d650f5cd344f86377ebbbaef5b19a0af3ae08101d1697f5656a52193000000000071c6f40024c342868a0c2a201b1b26a5d52c5d2f92a106c19ff926deb3fba1e74a444ecee3f8f507c062b949a2eaadd442b049417f82e8811526fa83c6d099d75323e068ffeca9dcd163761000c65d21dede72787ac350f25bdd3d29db6e9cb0e22c8124c724db33660c88784e2871b62ecf816846db7b469c71cad9a5dcfc5548ed2dd781006fa15b968facf4d79219646267eb187a670306d1ff1a59fc28ae00d36bb5a1cba659f48aa64a9022711a66105ef14401ff3948add265240aaad329ee76ba4c2300496746b86bcccacff5947c3fcb956cde2cffae10435960d7097f989aac742cf1047887f11584d20297958385e1715fe0f9b69141750c20d8134420eafec68fd10000000001555540006958aabfe4cc5d870e61fef82bcf1f2c3859e2bd8b1177e8a8872376b5cabace5dcb59b6fecada7e522d05f6f0e352939a6bfdf8c454fbe822cfa5ce97d17be0ffde44a4812cde9d04ec5c08dce6f9146586fdc8e081e05ec690b7effe24ea756f3d300f361203b61e1a39220c6eafa7852842674e317dcae5549c78c7144296ff004a6d0d2854c55e4c1de2f17dc4480b81652cfec37124ef41560a28c853482732434d1c006763b2e341528ae0bcc29fb76f1a4dafd99ade4fd75ec9cc9ca3f3d7001bcb6eb71e43eb22169ab721637551a8ec93838eb0825e9ecba9175297a00b146e9fdd244c5b722f29d3c46ec38840ba18f1f06ddec3dea844867386c2e1ac95"
            ),
            33554432,
            1024,
            2,
            z,
        )
        assert is_valid

    t2 = time.time()
    print(f"{z} IPS: {iters / (t2 - t1)}")
